/****************************************************************************

        Program    : server_acl.c
	Date       : Aug. 25 1997 -
	Input      : 2 IDs - 
	             User ID
		     Application ID
	output     : 

	programmer : Dongmin Kim (kimd@top.cis.syr.edu)
	Function   :
	     Read some informations from Application Flow Table,
	     use three tables (Resource, R_Constraint, T_Performance),
	     and generate Resource Allocation Table.

	Scheduling Algorithm :
	     Find machine(s) that makes the min of 

	          runtime (from T_Performance)
	          ----------------------------- X load_1 (from Resource)
	           load (from T_Performance)

	DB used in this program : mSQL (mini SQL)

	Host machine : emorgana.cat.syr.edu

	DB : VDCE, VDCE_TABLES

	To run this sample program, first the "msqld" should be running 
	on emorgana.cat.syr.edu.
	"msqld" is located on ~demo/vdce/mSQL/RS6K   /Minerva/bin/msqld
	                                      Solaris

*****************************************************************************/

#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <netdb.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include "msql.h"
#include <stdlib.h>

#define IP_length   30
#define name_length 20
#define id_length   20
#define Max_char    100

#define Tmp_HOST    "emorgana.cat.syr.edu"
#define DB_DEFAULT  "emorgana.cat.syr.edu"

#define DEMO        "1"
#define TEST_VDCE   "2"
#define ATM         "1"
#define TEST_CL     "4"

#define ANY         "0"
#define VDCE        "VDCE"
#define TABLES      "VDCE_TABLES"
#define NotUse      "n"
#define Ext_IO      "0"
#define Ext_Machine "***.***.***.***"
#define Master      "0"
#define Slave       "1"
#define Space       ' '
#define Spaces      " "

#define SUCCESS      1
#define FAIL         0
#define WAIT         1
#define DONE         0 
#define Parallel    "p"

#define TEST        0
#define PERFORMANCE 1
#define FAULT       2

#define Ready       "1111"
#define Error       "0000"
#define SELECT      "select"
#define NOTSELECT   "not"
#define Current_Domain 99999
#define CONT        10

typedef struct PORT_record {
  int  PortNo;
  char CommType[2];
  int  LinkTaskNo;
  int  LinkPortNo;
  char variable_name[Max_char];
} PORT;

typedef struct Machine_Type_record {
  char MachinePlatform[40];
  char MachineArch[20];
  char MachineOS[20];
  int  MachineNum;
} Machine;

typedef struct AFT_record {
  int  TaskNo;
  char ModuleID[id_length];
  char ModuleName[20];
  char ModuleType[2];
  int  InPort;
  PORT *inport;
  int  OutPort;
  PORT *outport;
  char CompType[2];
  char CommType[2];
  char ThreadType[2];
  char CommProtocol[2];
  char QoS[2048];
/**
  char MachineArch[20];
  char MachineOS[20];
  int  MachineNum;
**/
  int  No_Type;
  int  No_Machine;
  Machine *MachineInfo;
  char FileName[20];
  char Location[Max_char];
  char domain[20];
  char cluster[20];
} AFT;

typedef struct CIT_record {
  char v_name[Max_char];
  int  v_type;
  char s_name[Max_char];
  int  s_pid;
  int  s_port;
  int  s_tkid;
  int  s_socket;
  int  s_DataRep;
  char d_name[Max_char];
  int  d_pid;
  int  d_port;
  int  d_tkid;
  int  d_socket;
  int  d_DataRep;
} CIT;

typedef struct ACT_record {
  int  TaskNo;
  char ModuleID[id_length];
  int  MachineNum;
  char CompType[2];
  char CommType[2];
  char MachineType[10];
  char ThreadType[2];
  char CommProtocol[2];
  char QoS[2048];
  char MachineIP[15];
  char MachineHostname[60];
  char FileName[Max_char];
  char Role[2];
  int  DataRep;
} ACT;

typedef struct query_record {
  int  TaskNo;
  char IP[Max_char];
  char HostName[60];
  char Arch[20];
  char OS[20];
  float load_1;
  char directory[100];
  int  DataRep;
  int  total_process;
  char mem_real[20];
  char mem_free[20];
  char alias[20];
} Query_Result;

typedef struct domain_info {
  char DomainName[20];
  char RepIP[20];
  char PortNo[10];
  char dba[10];
  char directory[100];
  char JSVIP[20];
  int  no_cluster;
} DOMAIN;

typedef struct cluster_info {
  char ClusterName[20];
  char Cluster[20];
  char IP[20];
  char PortNo[10];
  char DBAName[10];
  char Traffic[20];
  char directory[100];
} CLUSTER;

typedef struct message_info {
  char header[56];
  char data[4];
} msg;

FILE   *fd;

AFT  *aft;
CIT  *cit;
ACT  *act;
Query_Result **Result_Table;
long seed;
DOMAIN *domain;
CLUSTER **cluster;

/**
char *server[] = {"ebeowulf.cat.syr.edu"};
char *server[] = {"corvus.dise.rl.af.mil"};
char *server[] = {"ebercilak.cat.syr.edu"};
**/

char DEFAULT_CLUSTER[name_length];
char DEFAULT_DOMAIN[name_length];

char My_Domain_Name[name_length];

int  *sockfd, *newsockfd;
int  *clustersockfd, *clusternewsockfd;
int  no_cluster;

int  chosen_cluster;
int  *lport;
int  *clusterlport;
int  start_cluster_index, end_cluster_index;
int  no_para;

/*
 * Function Name : pr_inet()
 *
 * Description   : convert internet address to ascii format(user readable)
 *
 * Input(s)      : internet - not user readable format
 *                 length   - length of internet address
 * 
 * Output(s)     : User readable internet address - ascii format
*/
pr_inet(listptr,length,addr)
char **listptr;
int  length;
char *addr;
{
 
     struct in_addr *ptr;
 
     while ( (ptr = (struct in_addr *) *listptr++) != NULL) 
       sprintf(addr,"%s",inet_ntoa(*ptr));
}

/***************************************************************
  Procedure : gethostaddr()
  Input     : 
  Output    : 
  Function  : 
****************************************************************/
char *gethostaddr() {

  char *hname;
  char *haddr;
  int  hlen=100;
  struct hostent *he;

  hname = (char *) calloc(hlen,sizeof(char));
  gethostname(hname,hlen);
  he = gethostbyname(hname);
  haddr = (char *) calloc(hlen,sizeof(char));

  switch ( he->h_addrtype) {
  case AF_INET:
    pr_inet(he->h_addr_list,he->h_length,haddr);
    break;
  default:
    break;
  }

  return(haddr);
}

/***************************************************************
  Procedure : SetDBconnection()
  Input     : database server IP
  Output    : connection status
  Function  : 
****************************************************************/
int SetDBconnection(db_IP)
char db_IP[30];
{

  int  sock;
  char action[1000];
  int  i, j, k;

  sock = -1;

  while( sock == -1 ) {

    /* connect to the host */
    sock = msqlConnect(db_IP);

    if ( sock == -1 ) {
      fprintf(fd, "Database is not available\n");
      fflush(fd);

      bzero(action, 1000);
      sprintf(action, "rsh %s -l demo /merlin2/users/demo/vdce/mSQL/Solaris/Minerva/bin/msqld &", db_IP);
      system(action);
    }
  }

  return sock;
}

/***************************************************************
  Procedure : CloseDBconnection()
  Input     : 
  Output    : 
  Function  : 
****************************************************************/
CloseDBconnection(sock)
int sock;
{

  /* close the DB */
  msqlClose(sock);
}

/***************************************************************
  Procedure : check_domain()
  Input     : 
  Output    : 
  Function  : 
****************************************************************/
check_domain(AES_serverIP, DB_serverIP)
char *AES_serverIP;
char *DB_serverIP;
{

  m_result *table;
  int  sock;
  char **result;
  int  nRow, nField;
  char action[1000];
  int  i, j, k;

  fprintf(fd, "Inside check the domain\n");
  fflush(fd);

  sock = SetDBconnection(DB_serverIP);

  /* open the VDCE DB  */
  msqlSelectDB(sock, VDCE);

  /* get domain information */
  bzero(action, 1000);
  sprintf(action, "select DomainName from DomainT where JSVIP = '%s' and  RepIP = '%s'",
	  AES_serverIP, DB_serverIP);
  msqlQuery(sock, action);

  /* store the result */
  table = msqlStoreResult();

  /* No. of rows in the result */
  nRow = msqlNumRows(table);

  msqlDataSeek(table, 0);
  result = msqlFetchRow(table);

  strcpy(My_Domain_Name, result[0]);

  strcpy(DEFAULT_DOMAIN, My_Domain_Name);

  if ( !strcmp(DEFAULT_DOMAIN, DEMO) ) {
    strcpy(DEFAULT_CLUSTER, ATM);
  }

  if ( !strcmp(DEFAULT_DOMAIN, TEST_VDCE) ) {
    strcpy(DEFAULT_CLUSTER, TEST_CL);
  }

  fprintf(fd, "Domain Name : %s      Rep. IP : %s\n", My_Domain_Name);
  fflush(fd);

  msqlFreeResult(table);

  CloseDBconnection(sock);
}

/***************************************************************
  Procedure : SetSockconnection()
  Input     : int socket_port, char *dest_IP
  Output    : 
  Function  : 
****************************************************************/
int SetSockconnection(socket_port, dest_IP)
int socket_port;
char *dest_IP;
{
  struct sockaddr_in serv_addr;
  struct hostent *host;
  long   tstart, tavg, tacc, at;
  int    sockfd2;
  int    msglen, rcvlen;
  int    i, socket_flag;
  int    optval = 1;
  int    optlen, buffsize = 2048;
  int    snd, rcv;
  int    NotSuccess = 1;

  fprintf(fd, "Inside connection\n");
  fprintf(fd, "server : %s\n", dest_IP);
  fflush(fd);

  while(NotSuccess) {

    /* Open a TCP socket (an Internet stream socket) */
    if ( ( sockfd2 = socket(AF_INET, SOCK_STREAM, 0) ) < 0 ) {
      fprintf(fd, "client: can't open stream socket.\n");
    }
    else { /* success to open a TCP socket */
      /* Specify the server address to connect. */
      bzero((char *) &serv_addr, sizeof(serv_addr));    /* Initialize  */
      serv_addr.sin_family      = AF_INET;              /* Internet    */
      serv_addr.sin_port        = htons(socket_port);   /* Port number */
      if ( ( host = gethostbyname(dest_IP) ) == 0 ) {
	fprintf(fd, "client: can`t get host address by name.\n");
      } 
      else { /* success to get the hostname */
	bcopy(host->h_addr, (caddr_t) &serv_addr.sin_addr, host->h_length);

	/* Connect to the server. */
	if (connect(sockfd2,(struct sockaddr *)&serv_addr,sizeof(serv_addr)) < 0) {
	  fprintf(fd, "client: can't connect to the server.");
	  fflush(fd);
	}
	else { /* success to connect to the server */
	  NotSuccess = 0;
	}
      }
    }
  }

  setsockopt(sockfd2,IPPROTO_TCP,TCP_NODELAY,(char *)&optval,sizeof(optval));
  snd = rcv = buffsize; /* BSD limits this up to 52000 */
  setsockopt(sockfd2, SOL_SOCKET, SO_SNDBUF, (char *) &snd, sizeof(snd));
  setsockopt(sockfd2, SOL_SOCKET, SO_RCVBUF, (char *) &rcv, sizeof(rcv));
  
  optlen = sizeof(buffsize);
  if ( getsockopt(sockfd2, SOL_SOCKET, SO_SNDBUF, (char *) &buffsize,
		  &optlen ) < 0 )
    fprintf(fd, "***** SO_SNDBUF getsockopt error\n");
  return(sockfd2);
}

/***************************************************************
  Procedure : set_ApplInfo
  Input     : char *userid, char *applid,  char *db_IP, int flag
  Output    : 
  Function  : set the applicaiton in ApplInfo table to d.
****************************************************************/
set_ApplInfo(userid, applid, db_IP, flag)
char *userid;
char *applid;
char *db_IP;
int  flag;
{

  int  sock;
  char action[1000];

  /* connect to the host */
  sock = SetDBconnection(db_IP);

  if ( sock == -1 ) {
    fprintf(fd, "Database is not available\n");
    fflush(fd);
    error_report();
  }

  /* open the VDCE DB  */
  msqlSelectDB(sock, VDCE);

  /* get total # of task, external input & output from ApplInfo table */
  bzero(action, 1000);
  if ( flag == WAIT ) {
    sprintf(action, "update ApplInfo set Status = 'd' where UserID = '%s' and ApplID = '%s'", userid, applid);
  }
  else if ( flag == DONE ) {
    sprintf(action, "update ApplInfo set Status = 'r' where UserID = '%s' and ApplID = '%s'", userid, applid);
  }

  msqlQuery(sock, action);

  CloseDBconnection(sock);
}

/***************************************************************
  Procedure : read_AFT
  Input     : char *userid, char *applid
  Output    : 
  Function  : read AFT table from DB
****************************************************************/
int read_AFT(userid, applid, DB_serverIP, no_task, no_channel)
char *userid;
char *applid;
char *DB_serverIP;
int *no_task;
int *no_channel;
{

  m_result *table;
  int  sock;
  char **result;
  int  nRow, nField;
  char action[1000];
  int  i, j, k, l;
  char string_tmp[200];
  char *tmp;
  char tmp2[40];
  char LU[10], INV[10], MULT_V[10], MULT_M[10];
  int  check_domain;
  int  ex_input, ex_output;

  fprintf(fd, "ACL : Inside AFT Query\n");
  fflush(fd);

  for ( i = 0 ; i < 20 ; i++ ) {
    /* connect to the host */
    sock = SetDBconnection(DB_serverIP);

    /* open the VDCE DB  */
    msqlSelectDB(sock, VDCE);

    /* get total # of task, external input & output from ApplInfo table */
    bzero(action, 1000);
    sprintf(action, "select TaskNum, InputFNum, OutputFNum from ApplInfo where UserID = '%s' and ApplID = '%s'", userid, applid);
    msqlQuery(sock, action);

    /* store the result */
    table = msqlStoreResult();

    /* No. of rows in the result */
    nRow = msqlNumRows(table);

    if ( nRow == 1 ) {
      /* No. of fields in the result */
      nField = msqlNumFields(table);

      msqlDataSeek(table, 0);
      result = msqlFetchRow(table);

      *no_task   = atoi(result[0]);
      ex_input  = atoi(result[1]);
      ex_output = atoi(result[2]);

      fprintf(fd, "Task : %d   Ext. input : %d   Ext. output : %d\n",
	      *no_task, ex_input, ex_output);
      fflush(fd);
    }
    else if ( nRow > 1 ) {
      fprintf(fd, "More than 2 informations are found in ApplInfo table\n");
      fflush(fd);
//      error_report();
    }
    else if ( nRow < 1 ) {
      fprintf(fd, "No informations are found in ApplInfo table\n");
      fflush(fd);
//      error_report();
    }

    msqlFreeResult(table);

    /* open the TABLES DB  */
    msqlSelectDB(sock, TABLES);

    /* get all informations from AFT table */
    bzero(action, 1000);
    sprintf(action, "select * from AFT_%s where ApplID = '%s' and ModuleType = 'c'",userid, applid);
    fprintf(fd, "action from AFT : %s\n", action);
    fflush(fd);

    msqlQuery(sock, action);

    /* store the result */
    table = msqlStoreResult();

    /* No. of rows in the result */
    nRow = msqlNumRows(table);

    if ( nRow == *no_task ) {
      fprintf(fd, "AFT_%s has been read\n", userid);
      fflush(fd);
      /* close the DB */
//      CloseDBconnection(sock);
      break;
    }
    else {
      fprintf(fd, "The task # in ApplInfo table and AFT_%s does not equal\n", userid);
      fflush(fd);
//      error_report();
    }
    if ( i == 19 ) {
      error_report();
    }
    else {
      sleep(1000000);
    }
  }

  fprintf(fd, "Assign\n");
  fflush(fd);

  *no_channel = ex_output;

  /* read the AFT table */
  aft = (struct AFT_record *) malloc((*no_task) * sizeof(struct AFT_record));

  for ( i = 0 ; i < *no_task ; i++ ) {

    msqlDataSeek(table, i);
    result = msqlFetchRow(table);

    aft[i].TaskNo = atoi(result[1]);
    strcpy(aft[i].ModuleID, result[2]);
    strcpy(aft[i].ModuleType, result[3]);
    strcpy(aft[i].CompType, result[4]);
    strcpy(aft[i].CommType, result[5]);
    strcpy(aft[i].ThreadType, result[6]);
    strcpy(aft[i].CommProtocol, result[7]);
    strcpy(aft[i].QoS, result[8]);

    fprintf(fd, "1 Original string for machine type : **%s** \n ", result[9]);
    fflush(fd);

    /* copy the string to tmp removing space */
    k = 0;
    for ( j = 0 ; j < 200 ; j++ ) {
      if (result[9][j] != Space) {
	string_tmp[k] = result[9][j];
	k++;
      }
    }

    fprintf(fd, "Original string for machine type : **%s** \n ", string_tmp);
    fflush(fd);

    if ( strtok(string_tmp, "|") == NULL ) {
      fprintf(fd, "Machine Type error : No Machine is defined\n");
      fflush(fd);
      error_report();
    }

    aft[i].No_Type = 1;

    while(1) {
      if ( strtok(0, "|") == NULL ) {
	break;
      }
      aft[i].No_Type++;
    }

    fprintf(fd, "Machine Type : %d\n", aft[i].No_Type);
    fflush(fd);

    /* copy the string to tmp removing space */
    k = 0;
    for ( j = 0 ; j < 200 ; j++ ) {
      if (result[9][j] != Space) {
	string_tmp[k] = result[9][j];
	k++;
      }
    }

    aft[i].MachineInfo  = (struct Machine_Type_record *) malloc(aft[i].No_Type * sizeof(struct Machine_Type_record));

    tmp = strtok(string_tmp, "|");
    strcpy(aft[i].MachineInfo[0].MachinePlatform, tmp);

    for ( j = 1 ; j < aft[i].No_Type ; j++ ) {
      tmp = strtok(0, "|");
      strcpy(aft[i].MachineInfo[j].MachinePlatform, tmp);
    }

    /* copy the string to tmp removing space */
    k = 0;
    for ( j = 0 ; j < 200 ; j++ ) {
      if (result[9][j] != Space) {
	string_tmp[k] = result[10][j];
	k++;
      }
    }

    tmp = strtok(string_tmp, "|");
    aft[i].MachineInfo[0].MachineNum = atoi(tmp);
    aft[i].No_Machine = aft[i].MachineInfo[0].MachineNum;

    for ( j = 1 ; j < aft[i].No_Type ; j++ ) {
      tmp = strtok(0, "|");
      aft[i].MachineInfo[j].MachineNum = atoi(tmp);
      aft[i].No_Machine = aft[i].No_Machine 
	                + aft[i].MachineInfo[j].MachineNum;
    }

    fprintf(fd, "Total Machine : %d\n", aft[i].No_Machine);
    fflush(fd);

    fprintf(fd, "Requested Machine Information\n");
    for ( j = 0 ; j < aft[i].No_Type ; j++ ) {
      fprintf(fd, "       Machine Type : %s      # of Machine : %d\n",
	      aft[i].MachineInfo[j].MachinePlatform, 
	      aft[i].MachineInfo[j].MachineNum);
    }
    fprintf(fd, "-----------------------------------------\n");
    fflush(fd);

    strcpy(aft[i].domain, result[13]);
    if ( !strcmp(aft[i].domain, Spaces) ) {
      strcpy(aft[i].domain, Spaces);
    }
    strcpy(aft[i].cluster, result[14]);
  }
       
  msqlFreeResult(table);

  fprintf(fd, "Done - read body\n");
  fflush(fd);

  /* open the VDCE DB  */
  msqlSelectDB(sock, VDCE);

  /* read from ModuleInfo table */
  for ( i = 0 ; i < *no_task ; i++ ) {

    /* get # of input port, output port, filename and location
       from ModuleInfo table */
    bzero(action, 1000);
    sprintf(action, "select InPort, OutPort, FileName, Location, Name from ModuleInfo where ID = '%s'", aft[i].ModuleID);

    fprintf(fd, "action : %s\n", action);
    fflush(fd);

    msqlQuery(sock, action);

    /* store the result */
    table = msqlStoreResult();

    /* No. of rows in the result */
    nRow = msqlNumRows(table);

    if ( nRow < 1 ) {
      fprintf(fd, "No informations are found in ModuleInfo table\n");
      fflush(fd);
      error_report();
    }
    msqlDataSeek(table, 0);
    result = msqlFetchRow(table);

    aft[i].InPort  = atoi(result[0]);
    aft[i].inport  = (struct PORT_record *) malloc(aft[i].InPort * sizeof(struct PORT_record));
    aft[i].OutPort = atoi(result[1]);
    aft[i].outport = (struct PORT_record *) malloc(aft[i].OutPort * sizeof(struct PORT_record));

    *no_channel = *no_channel + aft[i].InPort;

    fprintf(fd, "Task : %s    InPort : %d    OutPort : %d\n",
	   aft[i].ModuleID, aft[i].InPort, aft[i].OutPort);
    fflush(fd);

    strcpy(aft[i].FileName, result[2]);
/*    strcpy(aft[i].Location, result[3]); */
    strcpy(aft[i].Location, "/tmp/");
    strcpy(aft[i].ModuleName, result[4]);

    msqlFreeResult(table);
  }

  /* open the TABLES DB  */
  msqlSelectDB(sock, TABLES);

  /* get informations - input port */
  fprintf(fd, "action from Port - input\n");
  fflush(fd);

  for ( i = 0 ; i < *no_task ; i++ ) {

    /* get informations from Port table */
    bzero(action, 1000);
    sprintf(action, "select * from Port_%s where ApplID = '%s' and TaskNo = %d and PortType = 'i' order by PortNo", userid, applid, aft[i].TaskNo);
    msqlQuery(sock, action);

    fprintf(fd, "action from Port - input : %s\n", action);
    fflush(fd);

    /* store the result */
    table = msqlStoreResult();

    /* No. of rows in the result */
    nRow = msqlNumRows(table);

    if ( nRow != aft[i].InPort ) {
      fprintf(fd, "The input port # in AFT_%s table and Port_%s does not equal in task %d\n", userid, userid, aft[i].TaskNo);
      fflush(fd);
      error_report();
    }

    for ( j = 0 ; j < aft[i].InPort ; j++ ) {
      msqlDataSeek(table, j);
      result = msqlFetchRow(table);
      aft[i].inport[j].PortNo = atoi(result[3]);
      strcpy(aft[i].inport[j].CommType, result[4]);
      if ( !strcmp(aft[i].inport[j].CommType, NotUse) ) {
	*no_channel--;
      }
      if ( !strcmp(result[5], Ext_IO) ) {
	aft[i].inport[j].LinkTaskNo = -1;
	aft[i].inport[j].LinkPortNo = -1;
      }
      else {
	aft[i].inport[j].LinkTaskNo = atoi(result[5]);
	aft[i].inport[j].LinkPortNo = atoi(result[6]);
      }

      strcpy(aft[i].inport[j].variable_name, result[8]);
      strcat(aft[i].inport[j].variable_name, result[7]);
    }

    msqlFreeResult(table);
  }

  fprintf(fd, "Done - input port\n");
  fflush(fd);

  /* get informations - output port */
  fprintf(fd, "action from Port - output\n");
  fflush(fd);

  for ( i = 0 ; i < *no_task ; i++ ) {

    /* get informations from Port table */
    bzero(action, 1000);
    sprintf(action, "select * from Port_%s where ApplID = '%s' and TaskNo = %d and PortType = 'o' order by PortNo", userid, applid, aft[i].TaskNo);
    msqlQuery(sock, action);

    fprintf(fd, "action : %s\n", action);
    fflush(fd);

    /* store the result */
    table = msqlStoreResult();

    /* No. of rows in the result */
    nRow = msqlNumRows(table);

    if ( nRow != aft[i].OutPort ) {
      fprintf(fd, "The output port # in AFT_%s table and Port_%s does not equal\n", userid, userid);
      fprintf(fd, "From Port : %d    from ModuleInfo : %d\n", nRow, aft[i].OutPort);
      fflush(fd);
      error_report();
    }

    for ( j = 0 ; j < aft[i].OutPort ; j++ ) {
      msqlDataSeek(table, j);
      result = msqlFetchRow(table);
      aft[i].outport[j].PortNo = atoi(result[3]);
      strcpy(aft[i].outport[j].CommType, result[4]);
      if ( !strcmp(result[5], Ext_IO) ) {
	aft[i].outport[j].LinkTaskNo = -1;
	aft[i].outport[j].LinkPortNo = -1;
      }
      else {
	aft[i].outport[j].LinkTaskNo = atoi(result[5]);
	aft[i].outport[j].LinkPortNo = atoi(result[6]);
      }

      strcpy(aft[i].outport[j].variable_name, result[8]);
      strcat(aft[i].outport[j].variable_name, result[7]);
    }

    msqlFreeResult(table);
  }

  fprintf(fd, "Done - output port\n");
  fflush(fd);

  /* close the DB */
  CloseDBconnection(sock);

  for ( i = 0 ; i < *no_task ; i++ ) {
    for ( k = 0 ; k < aft[i].No_Type ; k++ ) {
      strcpy(aft[i].MachineInfo[k].MachineOS,
	     strtok(aft[i].MachineInfo[k].MachinePlatform, "/"));
      strcpy(aft[i].MachineInfo[k].MachineArch,
	     strtok(0, "/"));
    }
  }

  fprintf(fd, "inside read AFT   Task : %d    Channel : %d\n", *no_task, *no_channel);
  fflush(fd);
  return(*no_task);

}

/***************************************************************
  Procedure : print_AFT
  Input     : char *userid, char *applid, int no_task,
              int no_channel
  Output    : 
  Function  : 
****************************************************************/
print_AFT(userid, applid, no_task, no_channel)
char *userid;
char *applid;
int  no_task;
int  no_channel;
{

  int i, j;

  fprintf(fd, "AFT Information\n");
  fprintf(fd, "===========================================================\n");
  fprintf(fd, "User ID : %s     Application ID : %s\n", userid, applid);
  fprintf(fd, "Total # of Task : %d   Total # of channel : %d\n", no_task, no_channel);
  fflush(fd);


  for ( i = 0 ; i < no_task ; i++ ) {
    fprintf(fd, "-----------------------------------------------------------\n");
    fprintf(fd, "Task No : %d\n", aft[i].TaskNo);
    fprintf(fd, "Task ID : %s\n", aft[i].ModuleID);
    fprintf(fd, "Module Name : %s\n", aft[i].ModuleName);
    fprintf(fd, "Module Type : %s\n", aft[i].ModuleType);
    fprintf(fd, "Input Port No : %d\n", aft[i].InPort);
    fflush(fd);

    for ( j = 0 ; j < aft[i].InPort ; j++ ) {
      fprintf(fd, "       Port No : %d\n", aft[i].inport[j].PortNo);
      fprintf(fd, "       Comm Type : %s\n", aft[i].inport[j].CommType);
      fprintf(fd, "       Link Task No : %d\n", aft[i].inport[j].LinkTaskNo);
      fprintf(fd, "       Link Port No : %d\n", aft[i].inport[j].LinkPortNo);
      fprintf(fd, "       variable name : %s\n", aft[i].inport[j].variable_name);
    }
    fflush(fd);

    fprintf(fd, "Output Port No : %d\n", aft[i].OutPort);
    fflush(fd);

    for ( j = 0 ; j < aft[i].OutPort ; j++ ) {
      fprintf(fd, "       Port No : %d\n", aft[i].outport[j].PortNo);
      fprintf(fd, "       Comm Type : %s\n", aft[i].outport[j].CommType);
      fprintf(fd, "       Link Task No : %d\n", aft[i].outport[j].LinkTaskNo);
      fprintf(fd, "       Link Port No : %d\n", aft[i].outport[j].LinkPortNo);
      fprintf(fd, "       variable name : %s\n", aft[i].outport[j].variable_name);
    }
    fflush(fd);

    fprintf(fd, "Computation Type   : %s\n", aft[i].CompType);
    fprintf(fd, "Communication Type : %s\n", aft[i].CommType);
    fprintf(fd, "Communication Protocol : %s\n", aft[i].CommProtocol);

    fprintf(fd, "Machine Type  : %d\n", aft[i].No_Type);
    fprintf(fd, "Machine Total : %d\n", aft[i].No_Machine);
    fflush(fd);
    for ( j = 0 ; j < aft[i].No_Type ; j++ ) {
      fprintf(fd, "      Machine Plat : %s\n", aft[i].MachineInfo[j].MachinePlatform);
      fprintf(fd, "      Machine OS   : %s\n", aft[i].MachineInfo[j].MachineOS);
      fprintf(fd, "      Machine Arch : %s\n", aft[i].MachineInfo[j].MachineArch);
      fprintf(fd, "      Machine No   : %d\n", aft[i].MachineInfo[j].MachineNum);
    }

    fprintf(fd, "File Name          : %s\n", aft[i].FileName);
    fprintf(fd, "Location           : %s\n", aft[i].Location);
    fprintf(fd, "Domain             : %s\n", aft[i].domain);
    fprintf(fd, "cluster            : %s\n", aft[i].cluster);

    fflush(fd);
  }

  fprintf(fd, "===========================================================\n");
}

/***************************************************************
  Procedure : print_domain
  Input     : char *userid, char *applid
  Output    : 
  Function  : 
****************************************************************/
print_domain(userid, applid, no_domain)
char *userid;
char *applid;
int  no_domain;
{

  int i, j;

  fprintf(fd, "Domain Information   no domain : %d\n", no_domain);
  fprintf(fd, "===========================================================\n");
  fprintf(fd, "User ID : %s     Application ID : %s\n", userid, applid);
  fflush(fd);

  for ( i = 0 ; i < no_domain ; i++ ) {
    fprintf(fd, "-----------------------------------------------------------\n");
    fprintf(fd, "Domain Name : %s \n", domain[i].DomainName);
    fprintf(fd, "Server IP   : %s \n", domain[i].JSVIP);
    fprintf(fd, "Rep. IP     : %s \n", domain[i].RepIP);
    fprintf(fd, "directory   : %s \n", domain[i].directory);

    fflush(fd);
  }

  fprintf(fd, "===========================================================\n");
}

/***************************************************************
  Procedure : print_cluster
  Input     : 
  Output    : 
  Function  : 
****************************************************************/
print_cluster(no_domain)
int no_domain;
{

  int i, j;

  fprintf(fd, "Cluster Information\n");
  fprintf(fd, "===========================================================\n");

  for ( i = 0 ; i < no_domain ; i++ ) {
    fprintf(fd, "Domain Name : %s \n", domain[i].DomainName);
    for ( j = 0 ; j < domain[i].no_cluster ; j++ ) {
      fprintf(fd, "-----------------------------------------------------------\n");
      fprintf(fd, "Cluster Name : %s \n", cluster[i][j].ClusterName);
      fprintf(fd, "Cluster      : %s \n", cluster[i][j].Cluster);
      fprintf(fd, "IP           : %s \n", cluster[i][j].IP);
      fprintf(fd, "PortNo       : %s \n", cluster[i][j].PortNo);
      fprintf(fd, "DBAName      : %s \n", cluster[i][j].DBAName);
      fprintf(fd, "Traffic      : %s \n", cluster[i][j].Traffic);
      fprintf(fd, "directory    : %s \n", cluster[i][j].directory);

      fflush(fd);
    }
  }
  fprintf(fd, "===========================================================\n");
}

/***************************************************************
  Procedure : ready_report()
  Input     : 
  Output    : 
  Function  : 
****************************************************************/
ready_report(flag, socketfd)
int flag;
int socketfd;
{

  char header[56];
  char data[4];
  char message[62];

  bzero(header, 56);
  bzero(message, 62);

  strcpy(header, "VDCE");
  strcpy(&header[4], "3000");
/**
  strcpy(&header[8], "*********1*********2*********3*********4");
**/
  strcpy(&header[8], "                                        ");
  strcpy(&header[48], "60      ");
  strcpy(data, "1111");
  
  sprintf(message, "%s\n%s",header, data);
    
  fprintf(fd, "message : %s\n", message);
  fflush(fd);

  if ( writen(socketfd, message, sizeof(message)) < 0 ) {
    fprintf(fd, "acl: can't send size.\n");
    exit(1);
  }

}

/***************************************************************
  Procedure : CloseSockconnection()
  Input     : 
  Output    : 
  Function  : 
****************************************************************/
CloseSockconnection(socketfd)
int socketfd;
{

  close(socketfd);

}

/***************************************************************
  Procedure : error_report()
  Input     : 
  Output    : 
  Function  : 
****************************************************************/
error_report(socketfd)
int socketfd;
{

  char header[56];
  char data[4];
  char message[60];

  bzero(header, 56);
  bzero(message, 60);

  strcpy(header, "VDCE");
  strcpy(&header[4], "3000");
/**
  strcpy(&header[8], "*********1*********2*********3*********4");
**/
  strcpy(&header[8], "                                        ");
  strcpy(&header[48], "60      ");
  strcpy(data, Error);

  sprintf(message, "%s%s",header, data);
/**
  strcat(message, header);
  strcat(message, data);
**/
  fprintf(fd, "error message : %s\n", message);

  if ( writen(socketfd, message, sizeof(message)) < 0 ) {
    fprintf(fd, "acl: can't send size.\n");
    exit(1);
  }

  CloseSockconnection(socketfd);

}

/***************************************************************
  Procedure : clean_n_create_table
  Input     : char *userid, char *applid, int no_task, char *db_IP
  Output    : 
  Function  : 
****************************************************************/
clean_n_create_table(userid, applid, no_task, db_IP)
char *userid;
char *applid;
int  no_task;
char *db_IP;
{

  int  sock, i;
  char action[1000];

  fprintf(fd, "ACL : clean and create CIT, ACT, and time\n");
  fflush(fd);

  /* connect to the host */
  sock = SetDBconnection(db_IP);

  /* open the TABLES DB  */
  msqlSelectDB(sock, TABLES);

  /* delete CIT table */
  bzero(action, 1000);
  sprintf(action, "drop table CIT_%s_%s", userid, applid);
  msqlQuery(sock, action);

  /* create CIT table */
  bzero(action, 1000);
  sprintf(action, "create table CIT_%s_%s (
  TotalChannel INT,
  variable_name CHAR(100) NOT NULL,
  variable_type CHAR(1) NOT NULL,
  source_IP CHAR(15) NOT NULL,
  source_TaskNo INT,
  source_PortNo INT,
  source_LPort INT,
  source_socket INT,
  source_data_rep INT,
  dest_IP CHAR(15) NOT NULL,
  dest_TaskNo INT,
  dest_PortNo INT,
  dest_LPort INT,
  dest_socket INT,
  dest_data_rep INT)", userid, applid);

  msqlQuery(sock, action);

  /* delete ACT table */
  bzero(action, 1000);
  sprintf(action, "drop table ACT_%s_%s", userid, applid);
  msqlQuery(sock, action);

  /* create ACT table */
  bzero(action, 1000);
  sprintf(action, "create table ACT_%s_%s (
  TotalTask INT,
  TaskNo int,
  ModuleID CHAR(20) NOT NULL,
  MachineNum INT,
  CompType CHAR(1) NOT NULL,
  CommType CHAR(1) NOT NULL,
  MachineArch CHAR(20) NOT NULL,
  MachineOS   CHAR(20) NOT NULL,
  ThreadType CHAR(1),
  CommProtocol CHAR(1),
  QoS CHAR(2048),
  MachineIP CHAR(15) NOT NULL,
  MachineHostname CHAR(60) NOT NULL,
  FileName CHAR(100) NOT NULL,
  Role CHAR(1) NOT NULL,
  data_represent INT,
  load_1 REAL,
  total_process INT,
  memory_real CHAR(20),
  memory_free CHAR(20),
  alias CHAR(20))", userid, applid);

  msqlQuery(sock, action);

  /* delete time table */
  bzero(action, 1000);
  sprintf(action, "drop table time_%s_%s", userid, applid);
  msqlQuery(sock, action);

  /* create time table */
  bzero(action, 1000);
  sprintf(action, "create table time_%s_%s (
  TaskNo int PRIMARY KEY,
  TotalTask INT,
  ModuleName char(20) not null,
  ExeTime char(10) not null,
  Unit char(10) not null)", userid, applid);

  msqlQuery(sock, action);

  /* insert the task name and setup time */
  for ( i = 0 ; i < no_task ; i++ ) {
    bzero(action, 1000);
    sprintf(action, "insert into time_%s_%s values (%d, %d, '%s', '0', '2')",
	    userid, applid,
	    aft[i].TaskNo, no_task, aft[i].ModuleName);
    msqlQuery(sock, action);
    fprintf(fd, "%d  action : %s\n", i, action);
  }

  bzero(action, 1000);
  sprintf(action, "insert into time_%s_%s values (-1, %d, 'Setup', '0', '0')",
	  userid, applid, no_task);
  msqlQuery(sock, action);
  fprintf(fd, "action : %s\n",action);

  CloseDBconnection(sock);

}

/***************************************************************
  Procedure : machine_select()
  Input     : char *userid, char *applid
              char *db_IP, char *clusterDB, int task
  Output    : 
  Function  : 
****************************************************************/
int machine_select(userid, applid, db_IP, clusterDB, task)
char *userid;
char *applid;
char *db_IP;
char *clusterDB;
int task;
{

  m_result *table;
  int  sock;
  char **result;
  int  nRow, nField;
  char action[1000];
  int  i, j, k;
  int  start_index, end_index;
  int  found;
  char Arch[20], OS[20];

  fprintf(fd, "===========================================================\n");
  fprintf(fd, "ACL : Inside machine select\n");
  fflush(fd);

  sock = SetDBconnection(db_IP);

  /* open the selected cluster DB  */
  msqlSelectDB(sock, clusterDB);

  end_index = 0;

  for ( k = 0 ; k < aft[task].No_Type ; k++ ) {

    fprintf(fd, "Task : %d     No_Type : %d\n", task, aft[task].No_Type);
    fflush(fd);

    bzero(action, 1000);

    if ( !strcmp(aft[task].CompType, Parallel)) {
      sprintf(action, "select Resource.IP, Resource.HostName, Resource.Arch, Resource.OS, Resource.load_1, R_Constraint.directory, Resource.alive_check, Resource.DataRep, Resource.total_process, Resource.memory_real, Resource.memory_free, Resource.alias from Resource, R_Constraint where Resource.IP = R_Constraint.IP and Resource.alive = 'alive' and Resource.Arch = '%s' and Resource.OS = '%s' and R_Constraint.CompType = '%s' and R_Constraint.CommType = '%s' and R_Constraint.ModuleID = '%s' and Resource.CommProtocol = '%s' order by Resource.load_1",
	      aft[task].MachineInfo[k].MachineArch,
	      aft[task].MachineInfo[k].MachineOS,
	      aft[task].CompType, aft[task].CommType, aft[task].ModuleID,
	      aft[task].CommProtocol);
    }
    else {
      sprintf(action, "select Resource.IP, Resource.HostName, Resource.Arch, Resource.OS, Resource.load_1, R_Constraint.directory, Resource.alive_check, Resource.DataRep, Resource.total_process, Resource.memory_real, Resource.memory_free, Resource.alias from Resource, R_Constraint where Resource.IP = R_Constraint.IP and Resource.alive = 'alive' and Resource.Arch = '%s' and Resource.OS = '%s' and R_Constraint.CompType = '%s' and R_Constraint.ModuleID = '%s' and Resource.CommProtocol = '%s' order by Resource.load_1",
	      aft[task].MachineInfo[k].MachineArch,
	      aft[task].MachineInfo[k].MachineOS,
	      aft[task].CompType, aft[task].ModuleID,
	      aft[task].CommProtocol);
    }

    fprintf(fd, "action : %s\n", action);
    fprintf(fd, "Arch: %s   OS: %s\n", Arch, OS);
    fflush(fd);

    /* get informations from Resource and R_Constraint table */
    msqlQuery(sock, action);

    /* store the result */
    table = msqlStoreResult();

    /* No. of rows in the result */
    nRow = msqlNumRows(table);

    if ( nRow <  aft[task].MachineInfo[k].MachineNum ) {
      fprintf(fd, "Not enough machine User : %s  Application : %s   Task : %d   Found %d but %d needed\n",
	      userid, applid, aft[task].TaskNo, nRow,
	      aft[task].MachineInfo[k].MachineNum);
      fflush(fd);
      return(FAIL);
    }

    start_index = end_index;
    end_index = end_index + aft[task].MachineInfo[k].MachineNum;
    found = 0;

    seed = timer_elapsed();
    srand(seed);

    found = rand() % nRow;

    fprintf(fd, "Start : %d    End : %d   found : %d\n", start_index, end_index, found);
    fflush(fd);

    for ( j = start_index ; j < end_index ; j++ ) {

      if ( found >= nRow ) {
	found = 0;
      }

      msqlDataSeek(table, found);
      found++;
      result = msqlFetchRow(table);
      Result_Table[task][j].TaskNo = aft[task].TaskNo;
      strcpy(Result_Table[task][j].IP, result[0]);
      strcpy(Result_Table[task][j].HostName, result[1]);
      strcpy(Result_Table[task][j].Arch, result[2]);
      strcpy(Result_Table[task][j].OS, result[3]);
      fprintf(fd, "result[1] : %s\n", result[1]);
      fprintf(fd, "result[2] : %s\n", result[2]);
      fprintf(fd, "result[3] : %s\n", result[3]);
      fprintf(fd, "result[4] : %s    %f\n", result[4], atof(result[4]));
      fprintf(fd, "result[5] : %s\n", result[5]);
      fprintf(fd, "result[6] : %s\n", result[6]);
      fprintf(fd, "result[7] : %s\n", result[7]);
      fflush(fd);
      Result_Table[task][j].load_1 = atof(result[4]);
      strcpy(Result_Table[task][j].directory, result[5]);
      Result_Table[task][j].DataRep = atoi(result[7]);
      Result_Table[task][j].total_process = atoi(result[8]);
      strcpy(Result_Table[task][j].mem_real, result[9]);
      strcpy(Result_Table[task][j].mem_free, result[10]);
      strcpy(Result_Table[task][j].alias, result[11]);
    }

    msqlFreeResult(table);
    fprintf(fd, "check point end of task : %d and type : %d\n", i, k);
    fflush(fd);

  }

  seed = seed + atoi(userid);

  /* close the DB */
  CloseDBconnection(sock);

  return(SUCCESS);
}

/***************************************************************
  Procedure : create_CIT
  Input     : char *userid, char *applid, int no_task, 
              int no_channel, char *db_IP
  Output    : 
  Function  : 
****************************************************************/
create_CIT(userid, applid, no_task, no_channel, db_IP)
char *userid;
char *applid;
int  no_task;
int  no_channel;
char *db_IP;
{

  m_result *table;
  int  sock;
  char **result;
  int  nRow, nField;
  char action[1000];
  int  i, j, k;
  int  citport;
  char sourceIP[15], destIP[15];
  int  source_data_rep, dest_data_rep;

  fprintf(fd, "===========================================================\n");
  fprintf(fd, "ACL : Inside create CIT\n");
  fflush(fd);

  seed = timer_elapsed();
  srand(seed);

  citport = 9000 + rand() % 80000; 

  fprintf(fd, "seed : %d    Lport : %d\n", seed, citport);
  fflush(fd);

  /* connect to the host */
  sock = SetDBconnection(db_IP);

  if ( sock == -1 ) {
    fprintf(fd, "Database is not available\n");
    fflush(fd);
    error_report();
  }

  /* open the TABLES DB  */
  msqlSelectDB(sock, TABLES);

  for ( i = 0 ; i < no_task ; i++ ) {

    for ( j = 0 ; j < aft[i].InPort ; j++ ) {
      if ( strcmp(aft[i].inport[j].CommType, NotUse) ) {

	if ( !strcmp(aft[i].inport[j].CommType, Ext_IO) ) {
	  strcpy(sourceIP, Ext_Machine);
	}
	else {
	  /* source information */
	  bzero(action, 1000);
	  sprintf(action, "select MachineIP, data_represent from ACT_%s_%s where TaskNo = %d and Role = '%s'",
		  userid, applid,
		  aft[i].inport[j].LinkTaskNo, 
		  Master);

	  /* get informations from ACT */
	  msqlQuery(sock, action);

	  /* store the result */
	  table = msqlStoreResult();

	  /* No. of rows in the result */
	  nRow = msqlNumRows(table);

	  msqlDataSeek(table, 0);

	  result = msqlFetchRow(table);

	  strcpy(sourceIP, result[0]);
	  source_data_rep = atoi(result[1]);
	}

	/* destination information */
	bzero(action, 1000);
	sprintf(action, "select MachineIP, data_represent from ACT_%s_%s where TaskNo = %d and Role = '%s'",
		userid, applid,
		aft[i].TaskNo,
		Master);

	/* get informations from ACT */
	msqlQuery(sock, action);

	/* store the result */
	table = msqlStoreResult();

	/* No. of rows in the result */
	nRow = msqlNumRows(table);

	msqlDataSeek(table, 0);
	
	result = msqlFetchRow(table);

	strcpy(destIP, result[0]);
	dest_data_rep = atoi(result[1]);
/**
          for ( k = 0 ; k < no_task ; k++ ) {
	    if ( Result_Table[k][0].TaskNo == aft[i].inport[j].LinkTaskNo ) {
	      strcpy(sourceIP, Result_Table[k][0].IP);
	    }
	  }
	}

	for ( k = 0 ; k < no_task ; k++ ) {
	  if ( Result_Table[k][0].TaskNo == aft[i].TaskNo ) {
	    strcpy(destIP, Result_Table[k][0].IP);
	  }
	}
**/
	fprintf(fd, "task no : %d    input : %d    source IP : %s   dest IP : %s\n", aft[i].TaskNo, aft[i].inport[j].PortNo, sourceIP, destIP);

	bzero(action, 1000);
	citport = citport + 10;
	sprintf(action,
		"insert into CIT_%s_%s values(%d, '%s', '%s', '%s', %d, %d, %d, -1, %d, '%s', %d, %d, %d, -1, %d)",
		userid, applid, no_channel, 
		aft[i].inport[j].variable_name,	aft[i].inport[j].CommType,
		sourceIP, aft[i].inport[j].LinkTaskNo,
		aft[i].inport[j].LinkPortNo, citport, source_data_rep,
		destIP, aft[i].TaskNo, aft[i].inport[j].PortNo, citport,
		dest_data_rep);

	msqlQuery(sock, action);

      }
    }

    for ( j = 0 ; j < aft[i].OutPort ; j++ ) {
      if ( strcmp(aft[i].outport[j].CommType, NotUse) ) {

	if ( !strcmp(aft[i].outport[j].CommType, Ext_IO) ) {
	  strcpy(destIP, Ext_Machine);

	  /* source information */
	  bzero(action, 1000);
	  sprintf(action, "select MachineIP, data_represent from ACT_%s_%s where TaskNo = %d and Role = '%s'",
		  userid, applid,
		  aft[i].TaskNo,
		  Master);

	  /* get informations from ACT */
	  msqlQuery(sock, action);

	  /* store the result */
	  table = msqlStoreResult();

	  /* No. of rows in the result */
	  nRow = msqlNumRows(table);

	  msqlDataSeek(table, 0);

	  result = msqlFetchRow(table);

	  strcpy(sourceIP, result[0]);
	  source_data_rep = atoi(result[1]);

	  fprintf(fd, "task no : %d    output : %d    source IP : %s   dest IP : %s\n", aft[i].TaskNo, aft[i].outport[j].PortNo, sourceIP, destIP);

/**
	  for ( k = 0 ; k < no_task ; k++ ) {
	    if ( Result_Table[k][0].TaskNo == aft[i].TaskNo ) {
	      strcpy(sourceIP, Result_Table[k][0].IP);
	    }
	  }
**/	
	  bzero(action, 1000);
	  citport = citport + 10;
	  sprintf(action,
		  "insert into CIT_%s_%s values(%d, '%s', '%s', '%s', %d, %d, %d, -1, %d, '%s', %d, %d, %d, -1, %d)",
		  userid, applid, no_channel, 
		  aft[i].outport[j].variable_name, aft[i].outport[j].CommType,
		  sourceIP, aft[i].TaskNo, aft[i].outport[j].PortNo, citport,
		  source_data_rep,
		  destIP, aft[i].outport[j].LinkTaskNo,
		  aft[i].outport[j].LinkPortNo, citport,
		  dest_data_rep);

	  msqlQuery(sock, action);

	}
      }
    }
  }

  /* close the DB */
  CloseDBconnection(sock);

}

/***************************************************************
  Procedure : create_ACT
  Input     : char *userid, char *applid, int no_task, char *db_IP
  Output    : 
  Function  : 
****************************************************************/
create_ACT(userid, applid, no_task, db_IP)
char *userid;
char *applid;
int  no_task;
char *db_IP;
{

  int  sock;
  char action[1000];
  int  found;
  int  i, j, k;
  int  new_index;
  char role[2];
  char Arch[20], OS[20];

  fprintf(fd, "===========================================================\n");
  fprintf(fd, "ACL : Inside create ACT\n");
  fflush(fd);

  /* connect to the host */
  sock = SetDBconnection(db_IP);

  /* open the TABLES DB  */
  msqlSelectDB(sock, TABLES);


  for ( i = 0 ; i < no_task ; i++ ) {
    new_index = 0;
    for ( k = 0 ; k < aft[i].No_Type ; k++ ) {
      for ( j = 0 ; j < aft[i].MachineInfo[k].MachineNum ; j++ ) {
	if ( k == 0 && j == 0 ) {
	  strcpy(role, Master);
	}
	else {
	  strcpy(role, Slave);
	}

	bzero(action, 1000);
	sprintf(action, "insert into ACT_%s_%s values (%d, %d, '%s', %d, '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s',  '%s', '%s', '%s', %d, %6.3f, %d, '%s', '%s', '%s')",
		userid, applid,
		no_task, aft[i].TaskNo, aft[i].ModuleID,
		aft[i].No_Machine, aft[i].CompType, aft[i].CommType,
		aft[i].MachineInfo[k].MachineArch,
		aft[i].MachineInfo[k].MachineOS,
		aft[i].ThreadType,
		aft[i].CommProtocol, aft[i].QoS,
		Result_Table[i][new_index].IP,
		Result_Table[i][new_index].HostName,
		Result_Table[i][new_index].directory,
		role, 
		Result_Table[i][new_index].DataRep,
		Result_Table[i][new_index].load_1,
		Result_Table[i][new_index].total_process,
		Result_Table[i][new_index].mem_real,
		Result_Table[i][new_index].mem_free,
		Result_Table[i][new_index].alias);

	fprintf(fd, "action : %s\n", action);
	fprintf(fd, "Arch: %s    OS: %s\n", aft[i].MachineInfo[k].MachineArch,
		aft[i].MachineInfo[k].MachineOS);
	fflush(fd);

	msqlQuery(sock, action);
	new_index++;
      }
    }
  }

  /* close the DB */
  CloseDBconnection(sock);

}

/***************************************************************
  Procedure : update_ACT
  Input     : char *userid, char *applid, int no_task, char *db_IP
  Output    : 
  Function  : 
****************************************************************/
update_ACT(userid, applid, no_task, db_IP)
char *userid;
char *applid;
int no_task;
char *db_IP;
{

  m_result *table;
  char **result;
  int  sock;
  int  nRow, nField;
  char action[1000];
  int  found;
  int  i, j, k;
  int  new_index;
  char role[2];

  fprintf(fd, "===========================================================\n");
  fprintf(fd, "ACL : Inside update ACT\n");
  fflush(fd);

  /* connect to the host */
  sock = SetDBconnection(db_IP);

  if ( sock == -1 ) {
    fprintf(fd, "Database is not available\n");
    fflush(fd);
    error_report();
  }

  /* open the TABLES DB  */
  msqlSelectDB(sock, TABLES);

  bzero(action, 1000);

  sprintf(action, "select TaskNo from ACT_%s_%s where Role = '%s'",
	  userid, applid, Master);

  while(1) {
    fprintf(fd, "action : %s\n", action);
    fflush(fd);
    msqlQuery(sock, action);
  
    /* store the result */
    table = msqlStoreResult();

    /* No. of rows in the result */
    nRow = msqlNumRows(table);

    if ( nRow == no_task ) {
      break;
    }
  }

  bzero(action, 1000);
  sprintf(action, "update ACT_%s_%s set TotalTask = %d", userid, applid, no_task);
  msqlQuery(sock, action);

  /* close the DB */
  CloseDBconnection(sock);

}

/***************************************************************
  Procedure : read_DomainT()
  Input     : char *db_IP
  Output    : 
  Function  : 
****************************************************************/
int read_DomainT(db_IP)
char *db_IP;
{

  m_result *table;
  char **result;
  int  sock;
  int  nRow, nField;
  char action[1000];
  int  i, j, k;
  int  no_domain;

  fprintf(fd, "ACL : Inside read DomainT\n");
  fprintf(fd, "DB_serverIP : %s\n", db_IP);
  fprintf(fd, " My Domain Name : %s\n", My_Domain_Name);
  fflush(fd);

  /* connect to the host */
  sock = SetDBconnection(db_IP);

  /* open the TABLES DB  */
  msqlSelectDB(sock, VDCE);

  bzero(action, 1000);

  sprintf(action, "select DomainName, RepIP, RepPortNo, DBAName, ACLdirectory, JSVIP from DomainT");
  msqlQuery(sock, action);

  /* close the DB */
  msqlClose(sock);

  /* store the result */
  table = msqlStoreResult();

  /* No. of rows in the result */
  nRow = msqlNumRows(table);

  no_domain = nRow;
  domain = (struct domain_info *) malloc(nRow * sizeof(struct domain_info));

  fprintf(fd, "Candidate of Domain 1: %d\n", no_domain);
  fflush(fd);

  for ( i = 0 ; i < no_domain ; i++ ) {

    msqlDataSeek(table, i);
    result = msqlFetchRow(table);

    strcpy(domain[i].DomainName, result[0]);
    strcpy(domain[i].RepIP, result[1]);
    strcpy(domain[i].PortNo, result[2]);
    strcpy(domain[i].dba, result[3]);
    strcpy(domain[i].directory, result[4]);
    strcpy(domain[i].JSVIP, result[5]);

    fprintf(fd, "Domain Name : %s \n", domain[i].DomainName);
    fprintf(fd, "Server IP   : %s \n", domain[i].JSVIP);
    fprintf(fd, "Rep. IP     : %s \n", domain[i].RepIP);
    fprintf(fd, "DBA         : %s \n", domain[i].dba);
    fprintf(fd, "directory   : %s \n", domain[i].directory);
    fflush(fd);

  }

  msqlFreeResult(table);

  return(no_domain);
/**
  k = 0;
  for ( i = 0 ; i < no_domain ; i++ ) {
    if ( !strcmp(domain[i].DomainName, My_Domain_Name) ) {
      j = i;
      k = 1;
      break;
    }
  }

  if ( k == 1 ) {
    domain[j] = domain[no_domain - 1];
    no_domain--;
  }

  fprintf(fd, "Candidate of Domain 2: %d\n", no_domain);
  fflush(fd);

  sockfd    =  (int *) malloc(no_domain * sizeof(int ));
  newsockfd =  (int *) malloc(no_domain * sizeof(int ));
  lport     =  (int *) malloc(no_domain * sizeof(int ));
**/
}

/***************************************************************
  Procedure : read_ClusterT()
  Input     : char *db_IP, char *PortNo, char *DBAName, int index
  Output    : 
  Function  : 
****************************************************************/
read_ClusterT(db_IP, PortNo, DBAName, index)
char *db_IP;
char *PortNo;
char *DBAName;
int  index;
{

  m_result *table;
  char **result;
  int  sock;
  int  nRow, nField;
  char action[1000];
  int  i, j, k;
  int  no_cluster;

  fprintf(fd, "ACL : Inside read ClusterT\n");
  fprintf(fd, "DB_serverIP : %s\n", db_IP);
  fflush(fd);

  /* connect to the host */
  sock = SetDBconnection(db_IP);

  /* open the TABLES DB  */
  msqlSelectDB(sock, VDCE);

  bzero(action, 1000);

  sprintf(action, "select ClusterName, Cluster, IP, PortNo, DBAName, Traffic, directory from ClusterT");
  msqlQuery(sock, action);

  fprintf(fd, "action : %s\n", action);
  fflush(fd);

  /* store the result */
  table = msqlStoreResult();

  /* No. of rows in the result */
  nRow = msqlNumRows(table);

  no_cluster = nRow;

  domain[index].no_cluster = no_cluster;

  cluster[index] = 
    (struct cluster_info *) malloc(no_cluster * sizeof(struct cluster_info));

  fprintf(fd, "Candidate of Cluster: %d\n", no_cluster);
  fflush(fd);

  for ( i = 0 ; i < no_cluster ; i++ ) {

    msqlDataSeek(table, i);
    result = msqlFetchRow(table);

    strcpy(cluster[index][i].ClusterName, result[0]);
    strcpy(cluster[index][i].Cluster, result[1]);
    strcpy(cluster[index][i].IP, result[2]);
    strcpy(cluster[index][i].PortNo, result[3]);
    strcpy(cluster[index][i].DBAName, result[4]);
    strcpy(cluster[index][i].directory, result[5]);

    fprintf(fd, "Cluster Name : %s \n", cluster[index][i].ClusterName);
    fprintf(fd, "Cluster : %s \n", cluster[index][i].Cluster);
    fprintf(fd, "IP     : %s \n", cluster[index][i].IP);
    fprintf(fd, "DBAName     : %s \n", cluster[index][i].DBAName);
    fprintf(fd, "directory   : %s \n", cluster[index][i].directory);
    fflush(fd);
  }

  msqlFreeResult(table);

  /* close the DB */
  CloseDBconnection(sock);
}

/***************************************************************
  Procedure : domain_select()
  Input     : char *userid, char *applid, int no domain,
              int task, char *db_IP
  Output    : 
  Function  : 
****************************************************************/
int domain_select(userid, applid, no_domain, task, db_IP)
char *userid;
char *applid;
int  no_domain;
int  task;
char *db_IP;
{

  int i;
  int selected_domain;

  for ( i = 0 ; i < no_domain ; i++ ) {
    if ( !strcmp(aft[task].domain, domain[i].DomainName) ) {
      selected_domain = i;
      break;
    }
  }

  return(selected_domain);
}

/***************************************************************
  Procedure : cluster_select()
  Input     : char *userid, char *applid, int domain_index,
              int task, char *db_IP
  Output    : 
  Function  : 
****************************************************************/
int cluster_select(userid, applid, domain_index, task, db_IP)
char *userid;
char *applid;
int  domain_index;
int  task;
char *db_IP;
{

  int i;
  int selected_cluster;
  char domain_RepIP[IP_length];
  char domain_PortNo[name_length];
  char domain_dba[name_length];
  char clusterDB[name_length];
  int  status;

  if ( !strcmp(aft[task].cluster, ANY) ) {
    strcpy(aft[task].cluster, DEFAULT_CLUSTER);
  }

  for ( i = 0 ; i < domain[domain_index].no_cluster ; i++ ) {
    if ( !strcmp(aft[task].cluster, cluster[domain_index][i].ClusterName) ) {
      selected_cluster = i;
      break;
    }
  }

  strcpy(domain_RepIP, domain[domain_index].RepIP);
  strcpy(domain_PortNo, domain[domain_index].PortNo);
  strcpy(domain_dba, domain[domain_index].dba);
  strcpy(clusterDB, cluster[domain_index][selected_cluster].Cluster);

  status = machine_select(userid, applid, db_IP, clusterDB, task);

  return(status);
}

/***************************************************************
  Procedure : mapping()
  Input     : char *userid, char *applid, int no domain,
              int no task,int no channel, char *db_IP
  Output    : 
  Function  : 
****************************************************************/
mapping(userid, applid, no_domain, no_task, no_channel, db_IP)
char *userid;
char *applid;
int  no_domain;
int  no_task;
int  no_channel;
char *db_IP;
{

  int  i, j, k;
  int  index_task;
  int  index_selected_domain;
  int  status;

  /* setup the result table */
  Result_Table = (struct query_record **) malloc(no_task * sizeof(struct query_record *));

  for ( i = 0 ; i < no_task ; i++ ) {
    Result_Table[i] = (struct query_record *) malloc(aft[i].No_Machine * sizeof(struct query_record));
  }

  for ( index_task = 0 ; index_task < no_task ; index_task++ ) {
    if ( !strcmp(aft[index_task].domain, ANY) ) {
      strcpy(aft[index_task].domain, DEFAULT_DOMAIN);
    }
    index_selected_domain = 
      domain_select(userid, applid, no_domain, index_task, db_IP);
    status =
      cluster_select(userid, applid, index_selected_domain, index_task);

    if ( status == FAIL ) {
      set2default(userid, applid, index_task);
      status =
	cluster_select(userid, applid, index_selected_domain, index_task);
    }
  }
}

/***************************************************************
  Procedure : set2default()
  Input     : char *userid, char *applid, int task
  Output    : 
  Function  : 
****************************************************************/
set2default(userid, applid, task)
char *userid;
char *applid;
int  task;
{
  strcpy(aft[task].CompType, "s");
  strcpy(aft[task].CommType, "0" );
  strcpy(aft[task].CommProtocol, "0");
  aft[task].No_Machine = 1;
  strcpy(aft[task].MachineInfo[0].MachineArch, "Sparc");
  strcpy(aft[task].MachineInfo[0].MachineOS, "Solaris");
  aft[task].MachineInfo[0].MachineNum = 1;
  aft[task].No_Type = 1;
}

/***************************************************************
  Procedure : main()
  Input     : 
  Output    : 
  Function  : main function
****************************************************************/
main(argc, argv)
int  argc;
char **argv;
{

  char ch;
  char LOG[100];
  char userid[id_length], applid[id_length];
  int  taskno;
  int  i, j, k;
  int  socket_port;
  int  least_exe_time;
  int  flag;
  int  socketfd;

  char AES_serverIP[IP_length];
  char ACS_serverIP[IP_length];
  char DB_serverIP[IP_length];

  int  no_task;
  int  no_channel;
  int  no_domain;

  timer_reset();

  if ( argc < 8 ) {
    printf("Usage : (1) User ID\n");
    printf("Usage : (2) Application ID\n");
    printf("Usage : (3) socket number\n");
    printf("Usage : (4) AES Server host name\n");
    printf("Usage : (5) ACS Server host name\n");
    printf("Usage : (6) DB Server host name\n");
    printf("Usage : (7) flag\n");
    exit(1);
  }

  strcpy(userid, argv[1]);
  strcpy(applid, argv[2]);
  socket_port = atoi(argv[3]);
  strcpy(AES_serverIP, argv[4]);
  strcpy(ACS_serverIP, argv[5]);
  strcpy(DB_serverIP, argv[6]);
  flag = atoi(argv[7]);

  bzero(LOG, 100);

  fd = fopen("/tmp/acl_log", "w");

  fprintf(fd, "ACL started\n");
  fflush(fd);

  /* get the current acl's domain name and ip address */
  check_domain(AES_serverIP, DB_serverIP);

  close(fd);
  
  sprintf(LOG, "/tmp/%s_ACL_LOG_%s_%s", My_Domain_Name, userid, applid);

  fd = fopen(LOG, "w");

  if ( flag == PERFORMANCE || flag == FAULT ) {
    socketfd = SetSockconnection(socket_port, ACS_serverIP);
  }

  fprintf(fd, "ACL is running\n");
  fflush(fd);

  /* read AFT from DB */
  read_AFT(userid, applid, DB_serverIP, &no_task, &no_channel);
  fprintf(fd, "Between read and print   Task : %d    Channel : %d\n", no_task, no_channel);
  fflush(fd);
  print_AFT(userid, applid, no_task, no_channel);

  /* set the ApplInfo Table to developing status */
  set_ApplInfo(userid, applid, DB_serverIP, WAIT);

  /* delete and create tables - CIT, ACT and time table */
  clean_n_create_table(userid, applid, no_task, DB_serverIP);

  no_domain = read_DomainT(DB_serverIP);
  print_domain(userid, applid, no_domain);

  cluster = (struct cluster_info **) malloc(no_domain * sizeof(struct cluster_info *));

  for ( i = 0 ; i < no_domain ; i++ ) {
    read_ClusterT(domain[i].RepIP, domain[i].PortNo, domain[i].dba, i);
  }
  print_cluster(no_domain);

  mapping(userid, applid, no_domain, no_task, no_channel, DB_serverIP);

  create_ACT(userid, applid, no_task, DB_serverIP);
  update_ACT(userid, applid, no_task, DB_serverIP);
  create_CIT(userid, applid, no_task, no_channel, DB_serverIP);

  /* set the ApplInfo Table to ready status */
  set_ApplInfo(userid, applid, DB_serverIP, DONE);

  if ( flag == PERFORMANCE || flag == FAULT ) {
    ready_report(flag, socketfd);
    CloseSockconnection(socketfd);
  }
  fprintf(fd, "server acl done\n");
  fclose(fd);
}
